from bccd.gui.bccd import bccd
import os
# ~ from logging.handlers import RotatingFileHandler
import argparse
from textwrap import dedent

if __name__ == '__main__':

    # command line switches ---------------------------------------------------
    parser = argparse.ArgumentParser(description=dedent("""\
        Run B-NMR beamspot viewer."""),
        formatter_class=argparse.RawTextHelpFormatter)
    
    # logging level
    # ~ parser.add_argument("-d", "--debug",
                        # ~ help='Run in debug mode',
                        # ~ dest='debug',
                        # ~ action='store_true',
                        # ~ default=False)

    # parse
    args = parser.parse_args()

    # Setup logging -----------------------------------------------------------
    # ~ logger = logging.getLogger(logger_name)

    # get log filename
    # ~ try:
        # ~ filename = os.path.join(os.environ['HOME'],'.bfit.log')
    # ~ except KeyError:
        # ~ filename = 'bfit.log'

    # make handler
    # ~ handler = RotatingFileHandler(filename,
                                  # ~ mode='a',
                                  # ~ maxBytes=100*1000, # 100 kB max
                                  # ~ backupCount=1)

    # get level and format for output string
    # ~ if args.debug:
        # ~ level = logging.DEBUG
        # ~ fmt = '%(asctime)s %(levelname)-8s %(module)s.%(funcName)s() [%(lineno)d] -- %(message)s'
    # ~ else:
        # ~ level = logging.INFO
        # ~ fmt = '%(asctime)s %(levelname)-8s %(module)s -- %(message)s'
    
    # set
    # ~ handler.setFormatter(logging.Formatter(fmt))
    # ~ handler.setLevel(level)

    # ~ logger.addHandler(handler)
    # ~ logger.setLevel(level)
    
    # testing
    # ~ testfn = None
    # ~ def testfn(self):
        # ~ self.fetch_files.run.set("40123 40127 +40129")
        # ~ self.fetch_files.get_data()
        # ~ self.fit_files.populate()
        # ~ self.fit_files.do_fit()
        # ~ import matplotlib.pyplot as plt
        # ~ plt.close('all')
        # ~ self.notebook.select(2)
        
    # Run bccd ----------------------------------------------------------------
    bccd()




import os 

__all__ = ['gui','backend']
__version__ = '2.4.0'
__author__ = 'Derek Fujimoto'

from bccd.backend.fits import fits

logger_name = 'bccd'
icon_path = os.path.join(os.path.dirname(__file__),'images','icon.gif')






# Notebook tab specifying fits file object and drawing functions
# Derek Fujimoto
# Sept 2020

from tkinter import *
from tkinter import ttk, messagebox, filedialog

import os
import numpy as np
import pandas as pd

from bccd.backend.fits import fits
import bccd.backend.colors as colors
from bccd.backend.PltTracker import PltTracker as plt
from functools import partial

# =========================================================================== # 
class fits_tab(object):
    """
        Notebook tab specifying fits file object and drawing functions
        
        Data Fields:
            
            bccd: pointer to top level
            black: StringVar, black level
            entry_black: Entry widget for black value
            filename: name of .fits file
            id: id number for later deletion (key in bccd.tabs)
            input_names: dict, map input names to nice titles
            input_objs: dict of objects corresponding to input fields 
                             {input_name:(value,field,label)}
            img: fits image object
            old_alpha: int, last alpha draw value
            old_color: string, last color draw value
            plt: PltTracker obj, set to point at bccd.plt
            style: StringVar, drawing style
            styles: dict, map drawing style to draw function and input names
            tab_frame: tkk.Frame; top level frame for this tab
    """
    
    # (color, inverted)
    colours =  {'Greys':       True,
                'twilight':    True,
                'seismic':     False,
                'Reds':        True,
                'Blues':       True,
                'Greens':      True,
                'Purples':     True,
                'Oranges':     True,
                'viridis':     False,
                'plasma':      False,
                'inferno':     False,
                'magma':       False,
                'cividis':     False,
                'hot':         False,
                'afmhot':      False,
                'gist_heat':   False,
                'YlOrBr':      True,
                'YlOrRd':      True,
                'OrRd':        True,
                'PuRd':        True,
                'RdPu':        True,
                'BuPu':        True,
                'GnBu':        True,
                'PuBu':        True,
                'YlGnBu':      True,
                'PuBuGn':      True,
                'BuGn':        True,
                'YlGn':        True,  
                'binary':      True,
                'gist_yarg':   True,
                'gist_gray':   True,
                'gray':        False,
                'bone':        False,
                'pink':        False,
                'spring':      False,
                'summer':      False,
                'autumn':      False,
                'winter':      False,
                'cool':        True,
                'Wistia':      True,
                'copper':      False,
                'PiYG':        True,
                'PRGn':        True,
                'BrBG':        True,
                'PuOr':        True,
                'RdGy':        True,
                'RdBu':        True,
                'RdYlBu':      True,
                'RdYlGn':      True,
                'Spectral':    False,
                'coolwarm':    False,
                'bwr':         False,
                'twilight_shifted': False,
                'hsv':         True,
                }
    
    input_names = {'alpha':'Alpha (%): ',
                   'cmap':'Colour Map: ',
                   'nlevels':'Num. Contours: ',
                   'sigma':'Gaus. Filter Stdev.: ',
                   'imap':'Invert Colour Map'}
    
    # ======================================================================= #
    def __init__(self, bccd, tab_frame, filename, id):
        
        # inputs
        self.id = id
        self.bccd = bccd
        self.plt = bccd.plt
        self.tab_frame = tab_frame
        self.filename = filename
        self.old_color = tuple(self.colours.keys())[0]
        self.old_alpha = 100

        # read image
        img = fits(filename, plt=bccd.plt, rescale_pixels=bccd.rescale_pixels)
        self.img = img
        
        # variables
        self.black = StringVar()
        self.black.set(str(img.black))
        
        self.style = StringVar()
        
        # function, input names
        self.styles = { 'Greyscale':    (self.img.draw,'alpha','cmap','imap'),
                        'Contours':     (self.img.draw_contour,'alpha','nlevels','cmap','imap'),
                        'Gradient':     (self.img.draw_sobel,'alpha','cmap','imap'),
                        'Edges':        (self.img.draw_edges,'alpha','sigma','cmap','imap'),
                        }
                        
        # set draw style to previous
        if bccd.tabs:
            self.style.set(bccd.tabs[-1].style.get())
        else:
            self.style.set(list(self.styles.keys())[0])
        
        self.input_objs = {}
       
        # Column 0 -----------------------------------------------------------
        r = 0
        
        frame_column0 = ttk.Frame(tab_frame, relief='sunken',pad=5)
        frame_column0.grid(column=0,row=0,rowspan=10,sticky=(N,W),padx=5,pady=5)
        
        # show main title and exposure
        ttk.Label(frame_column0,
                  text=os.path.basename(filename)).grid(column=0,row=r,sticky=W); r+=1
        ttk.Label(frame_column0,
                  text='Exposure: %.3f s'%img.header['EXPOSURE']).grid(column=0,row=r,sticky=W); r+=1
        
        # show date and time, converted from utc to local
        date = img.datetime.strftime("%Y-%m-%d")
        time = img.datetime.strftime("%H:%M:%S")
        
        ttk.Label(frame_column0,text=date).grid(column=0,row=r,sticky=W); r+=1
        ttk.Label(frame_column0,text=time).grid(column=0,row=r,sticky=W); r+=1
        
        # Columnn 1 -----------------------------------------------------
        
        frame_column1 = ttk.Frame(tab_frame, relief='sunken',pad=5)
        frame_column1.grid(column=1,row=0,rowspan=10,sticky=(N,W),padx=5,pady=5)
        
        label_style = ttk.Label(frame_column1, text='Draw Style: ')
        self.combo_style = ttk.Combobox(frame_column1, textvariable=self.style, 
                                   state='readonly', width=20)
        self.combo_style['values'] = tuple(self.styles.keys())
        
        r = 0
        label_style.grid(column=0,row=r,sticky=W)
        self.combo_style.grid(column=1,row=r,sticky=W); r+=1
        
        # black level
        label_black = ttk.Label(frame_column1,text='Black Value: ')
        frame_black = ttk.Frame(frame_column1)
        self.entry_black = ttk.Entry(frame_black, textvariable=self.black, width=10)
        button_black = ttk.Button(frame_black, text='Reset', width=8, 
                                  command=self.reset_black)
        
        label_black.grid(column=0,row=r,sticky=W)
        frame_black.grid(column=1,row=r,sticky=W); r+=1
        self.entry_black.grid(column=0,row=0,sticky=W)
        button_black.grid(column=1,row=0,sticky=W, padx=2)
        
        # Inputs for draw style
        self.combo_style.bind("<<ComboboxSelected>>", 
                         partial(self.change_draw_fn, frame=frame_column1, row=r))
        r = self.input_place(frame_column1,r)+1
        
        # Draw buttons
        frame_draw = ttk.Frame(frame_column1)
        button_remove = ttk.Button(frame_draw,text='Remove',command=self.remove)
        button_draw_super = ttk.Button(frame_draw,text='Superimpose',command=self.draw)
        button_draw_new = ttk.Button(frame_draw,text='Draw New',command=self.draw_new)
        
    
        frame_draw.grid(column=0,row=r,sticky=(W,E,N),columnspan=2); r+=1
        frame_draw.columnconfigure(1,weight=1)
        c = 0
        button_remove.grid(column=c,row=0,sticky=W); c+=2
        button_draw_super.grid(column=c,row=0,sticky=E); c+=1
        button_draw_new.grid(column=c,row=0,sticky=E); c+=1
                
        button_close = ttk.Button(tab_frame,text='x',command=self.close, pad=0, 
                                  width=3)
        button_close.grid(column=10,row=0,sticky=(N,E),pady=5)
        
        # resizing
        tab_frame.grid_columnconfigure(9, weight=1)        # main area
        tab_frame.grid_rowconfigure(9,weight=1)            # main area
    
    # ======================================================================= #
    def close(self):
        """Remove the tab"""
        
        selected = self.bccd.notebook.select()
        self.bccd.notebook.forget(selected)
        self.tab_frame.destroy()
        
        for item in self.bccd.notebook.winfo_children():
            if str(item)==selected:
                item.destroy()       
                break
        
        for i,tab in enumerate(self.bccd.tabs):
            if tab is self:
                del self.bccd.tabs[i]
                break
    
    # ======================================================================= #
    def change_draw_fn(self, event, frame, row):
        """Change the draw function"""
        self.input_remove()
        self.input_place(frame, row)
    
    # ======================================================================= #
    def draw(self):
        """
            Draw image based on selection
        """
        
        # get draw style
        style = self.style.get()
        
        # get draw fn
        fn = self.styles[style][0]
        
        # set black level 
        self.img.set_black(float(self.black.get()))
        
        # get inputs
        options = {}
        
        for k,v in self.input_objs.items():
            if k in ('sigma',):
                options[k] = float(v[0].get())
            elif k == 'alpha':
                options[k] = v[0].get()/100
            else:
                options[k] = v[0].get()
        
        # draw
        fn(**options)
    
    # ======================================================================= #
    def draw_new(self):
        """
            Draw in a new window
        """
        self.plt.figure()
        self.draw()
        
    # ======================================================================= #
    def input_place(self, frame, row):
        """
            Create and grid new input fields
            
            frame:  ttk.Frame obj to draw in
            row:    starting row. Assume starting column is 0
            
            returns: final row
            adds items to self.input_objs
        """
        
        # setup
        style = self.style.get()
        input_names = self.styles[style][1:]
        
        # make objects
        for inpt in input_names:
            
            # alpha
            if inpt == 'alpha':
                
                # make elements
                label = ttk.Label(frame, text=self.input_names[inpt])
                value = IntVar()
                element = Spinbox(frame, textvariable=value, width=5, from_=0, to=100)
                value.set(self.old_alpha)
                
            # colour map
            elif inpt == 'cmap':
                
                # make elements
                label = ttk.Label(frame, text=self.input_names[inpt])
                value = StringVar()
                element = ttk.Combobox(frame, textvariable=value, 
                                       state='readonly', width=20)
                element.bind("<<ComboboxSelected>>", self.set_imap)
                element['values'] = tuple(self.colours.keys())
                value.set(self.old_color)
                
            # number contours
            elif inpt == 'nlevels':
                
                # make elements
                label = ttk.Label(frame, text=self.input_names[inpt])
                value = IntVar()
                element = Spinbox(frame, textvariable=value, width=5, from_=1, 
                                  to=1000)
                value.set(5)
                
            # invert colour map
            elif inpt == 'imap':
                
                # make elements
                label = ttk.Label(frame, text='')
                value = BooleanVar()
                element = ttk.Checkbutton(frame, text=self.input_names[inpt],
                        variable=value, onvalue=False, offvalue=True,
                        pad=5)
                value.set(self.colours[self.input_objs['cmap'][0].get()])
                
            # gaussian filter radius
            elif inpt == 'sigma':
                
                # make elements
                label = ttk.Label(frame, text=self.input_names[inpt])
                value = StringVar()
                element = ttk.Entry(frame, textvariable=value, width=10)
                element.insert(0,"1")
                
            else:
                raise RuntimeError('Input %s not implemented' % inpt)
                
            # grid
            label.grid(column=0, row=row, sticky=(N,W))
            element.grid(column=1, row=row, sticky=(N,W))
            
            # save 
            self.input_objs[inpt] = (value, element, label)
            
            # incrment the gridding row
            row += 1
                
        return row
        
    # ======================================================================= #
    def input_remove(self):
        """Remove draw function options"""
        
        self.old_color = self.input_objs['cmap'][0].get()
        
        try:
            self.old_alpha = self.input_objs['alpha'][0].get()
        except KeyError:
            pass
        
        for k,i in self.input_objs.items():
            i[1].grid_forget()
            i[2].grid_forget()
            i[1].update_idletasks()
            i[2].update_idletasks()
        
        keys = tuple(self.input_objs.keys())
        for k in keys:
            del self.input_objs[k]
            
        self.input_objs = {}
        
    # ======================================================================= #
    def remove(self):
        """
            Remove image from the active figure
        """
        self.plt._remove_drawn_object(self.plt.gca(),self.filename)
    
    # ======================================================================= #
    def reset_black(self):
        """
            Reset black to header value
        """
        black = self.img.header['BZERO']
        self.entry_black.delete(0,END)
        self.entry_black.insert(0,str(black))
        self.black.set(str(black))
        
    # ======================================================================= #
    def set_imap(self, *args):
        """
            Set imap based on color seclection
        """
        
        imap = self.input_objs['imap'][0]
        cmap = self.input_objs['cmap'][0]
        imap.set(self.colours[cmap.get()])



# Popup to set target settings
# Derek Fujimoto
# Sep 2020

from tkinter import *
from tkinter import ttk
import textwrap

from bccd.backend import colors
from bccd.backend.Target import Circle, Square, Rectangle, Ellipse


# ========================================================================== #
class popup_target(object):
    """
        Popup window for setting targeting options. 
        
        Data fields: 
            ax_list: list of axes on which the current target is drawn
            color: string. matplotlib color 
            frame_color: tk frame for showing the color of the target lines
            interactive: BooleanVar, if true, target can be interacted with, 
                         using draggablepoints
            radios: list of radio buttons
            result_label: ttk.label object for showing target info
            shape: StringVar, stores shape to draw
            target: Target object
            win: toplevel window
    """

    description = '\n'.join(textwrap.wrap("Draw a shape in the active figure, "+\
                                "synchronized across multiple figures.",width=30))
                    
    # colours in the default matplotlib cycle
    colors = {  'C0':'#1f77b4',
                'C1':'#ff7f0e', 
                'C2':'#2ca02c', 
                'C3':'#d62728',
                'C4':'#9467bd',
                'C5':'#8c564b', 
                'C6':'#e377c2',
                'C7':'#7f7f7f',
                'C8':'#bcbd22',
                'C9':'#17becf',
                }
                
    shapes = ('circle','square','rectangle','ellipse')

    # ====================================================================== #
    def __init__(self, bccd, color='C0'):
        self.bccd = bccd
        self.color = color
        self.ax_list = []
        
        # make a new window
        win = Toplevel(bccd.mainframe)
        win.title('Set and Draw Target')
        
        # target
        self.target = None
        
        # icon
        bccd.set_icon(win)
        
        # Key bindings
        win.bind('<Return>',self.draw)             
        win.bind('<KP_Enter>',self.draw)
    
        win.protocol("WM_DELETE_WINDOW", self.on_closing)
    
        # Menu 
        win.option_add('*tearOff', FALSE)
        menubar = Menu(win)
        win['menu'] = menubar
        self.interactive = BooleanVar()
        self.interactive.set(True)
        menubar.add_checkbutton(label="Interactive", variable=self.interactive,
                                selectcolor=colors.selected, 
                                command=self.hide_handles)
        
        # Column0 ----------------------------------------------------------
        frame_col0 = ttk.Frame(win, relief='sunken', pad=5)
        frame_col0.grid(column=0, row=0, sticky=(N,S,E,W), padx=5, pady=5)
    
        # Colour swatch
        self.frame_color = Frame(frame_col0, width=200, height=30)
        self.frame_color.grid(column=0,row=0,sticky=(N,W,S,E), padx=10, pady=10)
        self.frame_color.columnconfigure(0,weight=1)
        self.frame_color.rowconfigure(0,weight=1)
        self.set_frame_color(self.colors[self.color])
        
        # Header
        ttk.Label(frame_col0, text=self.description).grid(column=0, row=1,
                  sticky=(N,W), padx=10, pady=10)
        
        # target info label
        self.result_label = ttk.Label(frame_col0, text='')
        self.result_label.grid(column=0, row=2, sticky=(N,W), padx=10, pady=10)
        
        # Column1 ----------------------------------------------------------
        frame_col1 = ttk.Frame(win, relief='sunken', pad=5)
        frame_col1.grid(column=1, row=0, sticky=(N,S,E,W), padx=5, pady=5)
        
        # draw shape radio buttons
        self.shape = StringVar()
        self.radios = []
        for i,v in enumerate(self.shapes):
            rad = ttk.Radiobutton(frame_col1, 
                            text=v.title().rjust(max([len(s) for s in self.shapes])), 
                            variable=self.shape, 
                            value=v)
            rad.grid(column=0, row=i, sticky=(N,W))
            self.radios.append(rad)
            
        self.shape.set(self.shapes[0])
        
        # Row2 ----------------------------------------------------------
        
        frame_row2 = ttk.Frame(win, relief='sunken', pad=5)
        frame_row2.grid(column=0, row=1, sticky=(N,S,E,W), padx=5, pady=5, columnspan=2)
        frame_row2.columnconfigure(0,weight=1)
        frame_row2.columnconfigure(1,weight=1)
        
        # buttons
        button_draw = ttk.Button(frame_row2, text='Draw', command=self.draw)
        button_remove = ttk.Button(frame_row2, text='Remove', command=self.remove)
        
        button_draw.grid(column=1,   row=0, sticky=(N,E,W,S))
        button_remove.grid(column=0, row=0, sticky=(N,E,W,S))
        
        self.win = win
        
    # ====================================================================== #
    def draw(self,*args):
        """
            Add the target to the open figure
            Disable the radio buttons
            Create the target object
        """
        
        # disable the radio buttons
        for r in self.radios:
            r.config(state='disabled')
        
        if self.target is None:
            if self.shape.get() == 'circle':
                self.target = Circle(self, self.color, self.result_label, 250, 185, 50)  
            elif self.shape.get() == 'square':
                self.target = Square(self, self.color, self.result_label, 250, 185, 50)  
            elif self.shape.get() == 'rectangle':
                self.target = Rectangle(self, self.color, self.result_label, 250, 185, 50) 
            elif self.shape.get() == 'ellipse':
                self.target = Ellipse(self, self.color, self.result_label, 250, 185, 50, 50) 
            else: 
                raise RuntimeError("Undefined shape")
        
        self.target.draw(self.bccd.plt.gca())
    
        # set interactivity
        if not self.interactive.get():
            self.target.disable_drag_points()
            
    # ====================================================================== #
    def hide_handles(self):
        """
            Hide the interactive handles
        """
        
        try:
            if self.interactive.get():
                self.target.enable_drag_points()
            else:
                self.target.disable_drag_points()
        except AttributeError:
            self.interactive.set(not self.interactive.get())
        
    # ====================================================================== #
    def remove(self):
        """
            Remove the target from the open figure
        """
        self.target.remove(ax=self.bccd.plt.gca())
        
    # ====================================================================== #
    def set_frame_color(self, color='#FFFFFF'):
        """
            Set the color frame color
        """
        
        if color in self.colors.keys():
            self.frame_color.config(bg=self.colors[color])
        else:
            self.frame_color.config(bg=color)
    
    # ====================================================================== #
    def on_closing(self):
        try:
            self.target.remove_all()
        except AttributeError:
            pass
            
        self.bccd.targets.remove(self)
        self.win.destroy()



# Fits image object
# Derek Fujimoto
# Nov 2019

import os
import numpy as np
import pandas as pd

from bccd.backend.functions import gaussian2D
from astropy.io import fits as astrofits

from scipy.optimize import curve_fit
from scipy.integrate import dblquad

from bccd.backend.PltTracker import PltTracker
from matplotlib.patches import Circle

import skimage as ski
from skimage import filters
from skimage.feature import canny
from skimage.transform import hough_circle,hough_circle_peaks
from skimage.transform import probabilistic_hough_line
from skimage.transform import rescale

from datetime import datetime
from dateutil import tz

plt_global = PltTracker()

# =========================================================================== #
class fits(object):
    """
        Data Fields: 
        
            black:          float, pixel value corresponding to black (zero)
            data:           2D numpy array, pixel values
            data_original:  numpy array, pixel values
            datetime:       datetime object with the time the image was taken, 
                            in the local time zone
            filename:       name of the file
            header:         dict, header information
            
            mask:           (x,y,r) specifying circle to mask on
            
            plt:            PltTracker object or None
            result_center:      (par,names) fitting results
            result_cm:          (par,names) center of mass results
            result_fit2D:       (par,cov) fitting results
            result_gaussian2D:  (par,cov,names) fitting results
            result_gaussian2D_overlap: float, overlap
            
        Colormaps: 
            Greys
            Purples
            Yellows
            Blues
            Oranges
            Reds
            Greens
            ...
                
            https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
    """
    show_options = {'origin':'lower',
                    'interpolation':'nearest'}
    
    # ======================================================================= #
    def __init__(self,filename, plt=None, rescale_pixels=True):
        """
            Read the file
            self.plt: plot tracker
            rescale_pixels: if True, rescale image such that pixels are square
        """
        self.filename = filename
        self.read(filename,rescale_pixels=rescale_pixels)
        self.data_original = np.copy(self.data)
        self.set_mask(None)
        
        if plt is None:
            self.plt = plt_global
        else:
            self.plt = plt
            
    # ======================================================================= #
    def detect_lines(self,sigma=1,min_length=50,min_gap=3,theta=None,nlines=np.inf,
                     draw=True):
        """
            Detect lines in image
            
            nlines:     number of line s to find
            min_length: minimum length of lines to find
            min_gap:    minimum gap between pixels to avoid breaking the line    
            theta:      list of acceptable angles for the lines to point
            
            returns: list of points ((x0,y0),(x1,y1)) to identify the end points of 
                     the lines
        """
        
        # get raw data
        data = self.data
        
        # get edges
        edges = canny(data,sigma=sigma, low_threshold=0, high_threshold=1)
        
        # select lines
        lines = probabilistic_hough_line(edges,threshold=10,line_length=min_length,
                                         line_gap=min_gap,theta=theta)
        # draw
        if draw:
            self.plt.figure()
            self.plt.imshow(data,alpha=1,cmap='Greys_r',**self.show_options)
            edges = np.ma.masked_where(~edges,edges.astype(int))
            self.plt.imshow(edges,alpha=1,cmap='Reds_r',**self.show_options)
            
            for line in lines:
                self.plt.plot(*tuple(np.array(line).T))
                
        # return 
        return lines

    # ======================================================================= #
    def detect_hlines(self,sigma=1,min_length=50,min_gap=3,nlines=np.inf,
                   draw=True,**kwargs):
        """
            Detect horizontal lines in image
            
            nlines:     number of line s to find
            min_length: minimum length of lines to find
            min_gap:    minimum gap between pixels to avoid breaking the line    
            
            returns: list of y positions to identify each line
        """
        
        # make a set of ranges about pi/2
        theta = np.linspace(np.pi/2-0.01,np.pi/2+0.01,30)
        
        # get lines 
        lines = get_lines(sigma=sigma,
                          min_length=min_length,
                          min_gap=min_gap,
                          nlines=nlines,
                          draw=draw,
                          theta=theta,
                          **kwargs)
        
        # get y values of lines 
        return [l[0][1] for l in lines]
    
    # ======================================================================= #
    def detect_circles(self,rad_range,nlines=1,sigma=1,draw=True):
        """
            Detect circles in image
            
            rad_range:  specify raidus search range (lo,hi)
            alpha:      draw transparency
            nlines:     number of circles to find
            
            returns: (center_x,center_y,radius)
        """
        
        # get raw data
        data = np.copy(self.data)
        data[self.data.mask] = self.black
        
        # get edges
        edges = canny(data,sigma=sigma, low_threshold=0, high_threshold=1)
        
        # get radii
        hough_radii = np.arange(*rad_range, 2)
        hough_res = hough_circle(edges, hough_radii)
        
        # select cicles 
        accums, cx, cy, radii = hough_circle_peaks(hough_res, hough_radii,
                                                    total_num_peaks=nlines)
        
        # draw
        if draw:
            self.plt.imshow(self.filename,data,alpha=1,cmap='Greys_r',**self.show_options)
            edges = np.ma.masked_where(~edges,edges.astype(int))
            self.plt.imshow(self.filename+'edges',edges,alpha=1,cmap='Reds_r',**self.show_options)
            
            for center_y, center_x, radius in zip(cy, cx, radii):
                circle = Circle((center_x,center_y),radius,
                            facecolor='none',linewidth=1,edgecolor='g')
                self.plt.gca().add_patch(circle)
                
        # return 
        return (cx,cy,radii)
    
    # ======================================================================= #
    def draw(self,black=None,alpha=1,cmap='Greys',imap=True):
        """
            Draw fits file to matplotlib figure
            
            black:      set black level
            alpha:      draw transparency
            cmap:       colormap
            imap:       invert the colour map
        """
        
        # get raw data
        if black is not None:
            self.set_black(black)
        data = self.data
        
        # color map 
        if imap: cmap+='_r'
        
        # draw
        self.plt.imshow(id=self.filename,X=data,alpha=alpha,cmap=cmap,
                        info={  'style':'Greyscale', 
                                'black':self.black, 
                                'exposure_s':self.header['EXPOSURE'],
                                'date':self.datetime
                            },
                        **self.show_options)
    
    # ======================================================================= #    
    def draw_2Dfit(self,fn,*pars,levels=10,cmap='jet'):
        """
            Draw the fit function as contours
        
        """
        
        # get image shape
        shape = self.data.shape
        
        # get function image
        x = np.arange(shape[1])    
        y = np.arange(shape[0])    
        gauss = np.zeros((len(y),len(x)))
        for i in y:
            gauss[i-y[0],:] = fn(x,i,*pars)

        # draw image
        X,Y = np.meshgrid(x,y)
        ax = self.plt.gca()
        contours = ax.contour(X,Y,gauss,levels=levels,cmap=cmap)
        ax.clabel(contours,inline=True,fontsize='x-small',fmt='%g')
        
        self.contours = contours
        return contours

    # ======================================================================= #
    def draw_contour(self,nlevels=5,alpha=1,cmap='Greys',imap=True):
        """
            Draw contours of fits file to matplotlib figure
            
            nlevels:    number of contours to draw
            alpha:      draw transparency
            cmap:       colormap    
            imap:       invert the colour map
        """
        
        data = self.data
        
        # color map 
        if imap: cmap+='_r'
        
        # draw
        X,Y = np.meshgrid(*tuple(map(np.arange,data.shape[::-1])))
        ax = self.plt.gca()
        
        options = {k:val for k,val in self.show_options.items() if k != "interpolation"}
        self.plt.contour(self.filename,X,Y,data,levels=nlevels,cmap=cmap,alpha=alpha,
                         info={ 'style':'Contours', 
                                'black':self.black,
                                'exposure_s':self.header['EXPOSURE'],
                                'date':self.datetime
                            },
                         **options)
    
    # ======================================================================= #
    def draw_edges(self,sigma=1,alpha=1,cmap='Greys',imap=True):
        """
            Draw fits file to matplotlib figure
            
            sigma:      Standard deviation of the Gaussian filter.
            alpha:      draw transparency
            cmap:       colormap
            imap:       invert the colour map
        """
        
        # get raw data
        data = self.data
        
        # get edges
        data2 = np.copy(data)
        data2[data.mask] = self.black
        edges = canny(data2,sigma=sigma,low_threshold=0, high_threshold=1)
        
        # color map
        if imap: cmap += '_r'
        
        # draw
        edges = np.ma.masked_where(~edges,edges.astype(int))
        self.plt.imshow(self.filename, edges, alpha=alpha, cmap=cmap, 
                        info={  'style':'Edges', 
                                'black':self.black,
                                'exposure_s':self.header['EXPOSURE'],
                                'date':self.datetime
                            },
                        **self.show_options)
        
    # ======================================================================= #
    def draw_sobel(self,alpha=1,cmap='Greys',imap=False):
        """
            Draw fits file to matplotlib figure
            
            alpha:      draw transparency
            cmap:       colormap
            imap:       invert the colour map
        """
        
        # color map
        if imap: cmap += '_r'
        
        # draw
        data = np.copy(self.data)
        sbl = filters.sobel(data)
        sbl[self.data.mask]=0
        self.plt.imshow(self.filename, 
                        sbl,
                        alpha=alpha,
                        cmap=cmap,
                        info={  'style':'Gradient', 
                                'black':self.black,
                                'exposure_s':self.header['EXPOSURE'],
                                'date':self.datetime
                            },
                        **self.show_options)
    
        return sbl
        
    # ======================================================================= #
    def fit2D(self,function,**fitargs):
        """
            Fit general function to fits file
            
            function:   python function handle
            
            returns curve_fit output
        """
        
        # get data and trim the edges
        data = self.data[:300,:200]
        
        # flatten the image
        flat = np.ravel(data)
        
        # get number of fit parameters (first two are x,y)
        npar = function.__code__.co_argcount-2
        if 'p0' not in fitargs:
            fitargs['p0'] = np.ones(npar)
            
        # get zero
        zero = np.min(flat)
        flat -= zero
        
        # normalize
        flat /= np.max(flat)
        
        # flatten the funtion 
        def fitfn(xy,*pars):    
            output = function(*xy,*pars)
            return np.ravel(output)
        
        # fit
        x = np.indices(data.shape)[::-1]
        par,cov = curve_fit(fitfn,x,flat,**fitargs)
        
        self.result_fit2D = (par,cov)
        return (par,cov)
    
    # ======================================================================= #    
    def fit_gaussian2D(self,draw=True,**fitargs):
        """
            Fit 2D gaussian to image
            
            draw:       draw the output
        """
        
        # get data 
        data = self.data
        
        # estimate moments https://scipy-cookbook.readthedocs.io/items/FittingData.html
        total = data.sum()
        X, Y = np.indices(data.shape)
        x = (X*data).sum()/total
        y = (Y*data).sum()/total
        col = data[:, int(y)]
        width_x = np.sqrt(np.abs((np.arange(col.size)-y)**2*col).sum()/col.sum())
        row = data[int(x), :]
        width_y = np.sqrt(np.abs((np.arange(row.size)-x)**2*row).sum()/row.sum()) 
        
        # fit 
        p0 = (x,y,width_x,width_y,1,0)
        par,cov = self.fit2D(gaussian2D,p0=p0)
        std = np.diag(cov)**0.5
        
        # make output
        df = pd.DataFrame({"result":par,"error":std},
                            index=('x0','y0','sigmax','sigmay','amp','theta'))
        
        # draw output
        if draw:
            self.plt.figure()    
            self.draw()
            contours = self.draw_2Dfit(gaussian2D,*par[:4],1,0,**fitargs)
            
            self.plt.xlim((par[0]-4*par[2],par[0]+4*par[2]))
            self.plt.ylim((par[1]-4*par[3],par[1]+4*par[3]))
            self.plt.gca().clabel(contours,inline=True,fontsize='x-small',fmt='%g')
    
        self.result_gaussian2D = df
        
        return df
    
    # ======================================================================= #    
    def get_center(self,draw=True):
        """
            Get image center of mass
            
            draw:       if true, draw the output
            returns:    x0,y0,sigx,sigy
        """
        
        # get raw data
        data = self.data
        black = max(self.black,self.header['BZERO'])
            
        # compress
        sumx = np.ma.mean(data,axis=0)
        sumy = np.ma.mean(data,axis=1)
        
        # shift baseline
        sumx -= black
        sumy -= black
        
        # normalize
        normx = np.ma.max(sumx)
        normy = np.ma.max(sumy)
        
        sumx /= normx
        sumy /= normy
        
        # fit with gaussian
        gaus = lambda x,x0,sig,amp,base : amp*np.exp(-((x-x0)/(2*sig))**2)+base
        
        parx,cov = curve_fit(gaus,np.arange(len(sumx)),sumx,p0=(180,10,1,0),
                                bounds=((0,0,0,-np.inf),np.inf))
        stdx = np.diag(cov)**0.5
        
        pary,cov = curve_fit(gaus,np.arange(len(sumy)),sumy,p0=(260,10,1,0),
                                bounds=((0,0,0,-np.inf),np.inf))
        stdy = np.diag(cov)**0.5               
        
        # draw
        if draw:
            self.plt.figure()
            self.plt.plot(sumx*normx,label='x')
            self.plt.plot(sumy*normy,label='y')
            
            fitx = np.linspace(0,max(len(sumx),len(sumy)),5000)
            self.plt.plot(fitx,gaus(fitx,*parx)*normx,color='k')
            self.plt.plot(fitx,gaus(fitx,*pary)*normy,color='k')     
            self.plt.legend()
            
            self.plt.figure()
            self.plt.imshow(self.filename,data,cmap='Greys_r',**self.show_options)
            self.plt.errorbar(parx[0],pary[0],xerr=2*parx[1],yerr=2*pary[1],fmt='o',
                          fillstyle='none',markersize=9)
                          
            if pary[1] > 2 and parx[1] > 2:
                self.plt.ylim(pary[0]-pary[1]*6,pary[0]+pary[1]*6)   
                self.plt.xlim(parx[0]-parx[1]*6,parx[0]+parx[1]*6)
                
                
        self.result_center = ((parx[0],pary[0],parx[1],pary[1]),
                              ('x0','y0','sigx','sigy'))
                
        # return 
        return (parx[0],pary[0],parx[1],pary[1])

    # ======================================================================= #
    def get_cm(self,draw=True):
        """
            Get image center of mass
            
            draw:       if true, draw the output
        """
        
        # get raw data
        data = self.data
        
        # compress
        sumx = np.ma.mean(data,axis=0)
        sumy = np.ma.mean(data,axis=1)
        
        # estimate center with weighted average
        sumx -= np.ma.min(sumx)
        sumy -= np.ma.min(sumy)
        
        nsumx = len(sumx)
        nsumy = len(sumy)
        
        cx = np.ma.average(np.arange(nsumx),weights=sumx)
        cy = np.ma.average(np.arange(nsumy),weights=sumy)

        # draw
        if draw:
            self.plt.figure()
            self.plt.imshow(self.filename,data,cmap='Greys_r',**self.show_options)
            self.plt.plot(cx,cy,'x')
                
        self.result_cm = ((cx,cy),('x0','y0'))
        
        # return 
        return (cx,cy)

    # ======================================================================= #
    def get_gaussian2D_overlap(self,ylo,yhi,xlo,xhi):
        """
            Get integral of gaussian2D PDF within some interval, normalized to the 
            area such that the returned overlap is the event probability within the 
            range. 
            
            ylo:    lower integration bound [outer] (float)
            yhi:    upper integration bound [outer] (float)
            xlo:    lower integration bound [inner] (lambda function)
            xlhi:   upper integration bound [inner] (lambda function)
            
                integration is: 
                    
                    int_y int_x G(x,y) dx dy
            
            
            returns overlap as given by dblquad
        """
        
        # get fitting results 
        x0,y0,sx,sy,amp,theta = self.result_gaussian2D['result']
        
        # get normalized volume
        # https://en.wikipedia.org/wiki/Gaussian_function
        # ~ a = 0.5*(np.cos(theta)/sx)**2 + 0.5*(np.sin(theta)/sy)**2
        # ~ b = 0.25*-(np.sin(theta)/sx)**2 + 0.25*(np.sin(theta)/sy)**2
        # ~ c = 0.5*(np.sin(theta)/sx)**2 + 0.5*(np.cos(theta)/sy)**2
        # ~ amp = np.sqrt(a*c-b**2)/np.pi
        
        amp = 1/(2*np.pi*sx*sy)
        
        # make PDF
        gaus = lambda x,y: gaussian2D(x,y,x0,y0,sx,sy,amp,theta)
        
        # integrate: fraction of beam overlap
        overlap =  dblquad(gaus,ylo,yhi,xlo,xhi)[0]
        
        self.result_gaussian2D_overlap = overlap
        return overlap
        
    # ======================================================================= #
    def read(self,filename,rescale_pixels=True):
        """
            Get xy data from fits file. Values are brightness of pixel. 
            
            filename:       name of file to open
            rescale_pixels: if True, rescale image such that pixels are square
        """
        
        # open the file
        filename = os.path.join(os.getcwd(),filename)
        fid = astrofits.open(filename)[0]
        
        # read the header
        self.header = fid.header
        
        # get the data
        data = fid.data

        # fix bad pixels: set to max
        data[data<self.header['BZERO']] = np.max(data)
        self.black = self.header['BZERO']
        
        # rescale image to correct pixel size asymmetry
        if rescale_pixels:
            aspect = self.header['YPIXSZ']/self.header['XPIXSZ']
            
            # always enlarge image, never make it smaller
            if aspect > 1:      resc = (aspect,1)
            else:               resc = (1,1/aspect)
            
            data = rescale(data,resc,order=3,multichannel=False,preserve_range=True) 
        
        self.data = data
        
        # get the time and date
        date = self.header['DATE-OBS']
        utc = datetime.strptime(date, '%Y-%m-%dT%H:%M:%S')
        utc = utc.replace(tzinfo=tz.tzutc())
        self.datetime = utc.astimezone(tz=None)
        
    # ======================================================================= #
    def set_black(self,black):
        """
            Clean, remove lowest values
            
            black:  value to set to black, all pixels of lower value raised to 
                    this level
        """
        
        # set the black input
        self.black = black
        
        # apply black input and mask settings
        self.set_mask(self.mask)
        
    # ======================================================================= #
    def set_mask(self,mask):
        """
            Mask image data
            mask:       (x,y,r) specifying center and radius of circle to mask on
        """
        
        # set the mask input
        self.mask = mask
        
        # reset blacklevel
        data = np.copy(self.data_original)
        data[data<self.black] = self.black
        
        # masking
        if mask is not None:     
            window = np.ones(self.data.shape)
            rr,cc = ski.draw.disk((mask[1],mask[0]),mask[2],shape=data.shape)
            window[rr,cc] = 0
            data = np.ma.array(data,mask=window)
        
        # make as a masked array
        else:
            data = np.ma.asarray(data)
        
        self.data = data
        



# Target for targeting
# Derek Fujimoto
# Sep 2020 

import matplotlib.patches as patches
from functools import partial
import numpy as np

class Target(object):
    
    """
        Drawing shapes on lots of figures
        
        Data fields:
            ax_list: list of axis
            bccd: bccd object
            color: string, maplotlib color name for coloring everything
            figures: list of figures to update
            label: ttk label object to update text on properties
            popup_target: popup_target object
            points: list of all DraggablePoints
            
    """

    # ======================================================================= #
    def __init__(self, popup_target, color, label):
        
        self.bccd = popup_target.bccd 
        self.popup_target = popup_target 
        self.ax_list = []
        self.color = color
        self.label = label
        self.points = []
        
    # ======================================================================= #
    def disable_drag_points(self):
        """
            Prevent interaction with the draggable points
        """
        # reduce interaction region to zero
        for dragpt in self.points:
            for pt in dragpt.points: 
                try:
                    pt.set_pickradius(0)
                    pt.set_markersize(0)
                except Exception:
                    pass
        
    # ======================================================================= #
    def draw(self, ax):
        """
            Axis operations for drawing
        """
        
        # check if already drawin in these axes
        if ax in self.ax_list:
            raise RuntimeError("Axis already contains this target") 
                
        # connect with window close
        ax.figure.canvas.mpl_connect('close_event', partial(self.remove,ax=ax))
                
        # add axes to list
        self.ax_list.append(ax)
        
    # ======================================================================= #
    def enable_drag_points(self):
        """
            Enable interaction with the draggable points
        """
        # reset interaction region
        for dragpt in self.points:
            for pt in dragpt.points: 
                pt.set_pickradius(DraggablePoint.size)
                pt.set_markersize(DraggablePoint.size)
        
    # ======================================================================= #
    def remove(self, *args, ax=None):
        """
            Remove the DraggablePoints and patch from the axis
        """
        if ax not in self.ax_list:
            return
                
        # remove points 
        for pt in self.points:     
            pt.remove(ax)
            
        # remove patches
        for patch in ax.patches:
            if patch in self.patches:
                ax.patches.remove(patch)
                self.patches.remove(patch)
        
        # remove axis from list
        self.ax_list.remove(ax)
        
    # ======================================================================= #
    def remove_all(self):
        """
            Remove the DraggablePoints and patch from all axes
        """
        for ax in self.ax_list.copy():
            self.remove(ax=ax)
            
class Circle(Target):
    """
        Drawing circle target shapes on lots of figures
        
        Data fields:
            pt_center, pt_radius: DraggablePoints
            patches: mpl.patches for patches
            x,y: center coordinates
            r: radius
    """

    # ======================================================================= #
    def __init__(self, popup_target, color, label, x, y, r):
        
        super().__init__(popup_target, color, label)
        
        # save circle position
        self.x = x
        self.y = y
        self.r = r
        
        # place circle at the center of the window
        self.patches = []
        
        # center
        self.pt_center = DraggablePoint(self,self.update_center,
                            setx=True,sety=True,color=self.color, marker='x')
        
        # radius
        self.pt_radius = DraggablePoint(self,self.update_radius,
                            setx=True,sety=False,color=self.color, marker='o')
        
        self.points = [self.pt_center, self.pt_radius]
        self.update_popup_label()
        
    # ======================================================================= #
    def draw(self, ax):
        """Add the target to the current axes"""
        
        super().draw(ax)
        
        # draw things
        self.patches.append(patches.Circle((self.x,self.y),self.r,
                                     fill=False, 
                                     facecolor='none',
                                     lw=1,
                                     ls='-',
                                     edgecolor=self.color))
        ax.add_patch(self.patches[-1])
        self.pt_center.add_ax(ax, self.x, self.y)
        self.pt_radius.add_ax(ax, self.x+self.r, self.y,)

    # ======================================================================= #
    def update_popup_label(self):
        """Update popup window label with info on target"""
        
        self.label.config(text='x0 = %.3f\ny0 = %.3f\nr = %.3f' % (self.x, self.y, self.r))
        
    # ======================================================================= #
    def update_center(self, x, y):
        """
            Update circle position based on DraggablePoint
        """
        self.pt_radius.set_xdata(x+self.r)
        self.pt_radius.set_ydata(y)
        
        for c in self.patches:
            c.set_center((x,y))
        
        self.x = x
        self.y = y
        self.update_popup_label()
    
    # ======================================================================= #
    def update_radius(self, x, y):
        """
            Update circle radius based on DraggablePoint
        """

        self.r = abs(self.x-x)
        
        for c in self.patches:
            c.set_radius(self.r)
            
        self.update_popup_label()
    
class Square(Target):
    """
        Drawing square target shapes on lots of figures
        
        Data fields:
            pt_center, pt_side: DraggablePoints
            square: mpl.patches for patches
            x,y: center coordinates
            side: side length
    """

    # ======================================================================= #
    def __init__(self, popup_target, color, label, x, y, side):
        
        super().__init__(popup_target, color, label)
        
        # save circle position
        self.x = x
        self.y = y
        self.side = side
        
        # place circle at the center of the window
        self.patches = []
        
        # center
        self.pt_center = DraggablePoint(self,self.update_center,
                            setx=True,sety=True,color=self.color, marker='x')
        
        # radius
        self.pt_side = DraggablePoint(self,self.update_side,
                            setx=True,sety=False,color=self.color, marker='s')
        
        self.points = [self.pt_center, self.pt_side]
        self.update_popup_label()
        
    # ======================================================================= #
    def draw(self, ax):
        """Add the target to the current axes"""
        
        super().draw(ax)
        
        # draw things
        self.patches.append(patches.Rectangle((self.x-self.side,self.y-self.side),
                                            width=self.side*2,
                                            height=self.side*2,
                                            fill=False, 
                                            facecolor='none',
                                            lw=1,
                                            ls='-',
                                            edgecolor=self.color))
        ax.add_patch(self.patches[-1])
        self.pt_center.add_ax(ax, self.x, self.y)
        self.pt_side.add_ax(ax, self.x+self.side, self.y)

    # ======================================================================= #
    def update_popup_label(self):
        """Update popup window label with info on target"""
        
        self.label.config(text='x0 = %.3f\ny0 = %.3f\nside = %.3f' % (self.x, self.y, self.side*2))
        
    # ======================================================================= #
    def update_center(self, x, y):
        """
            Update circle position based on DraggablePoint
        """
        self.pt_side.set_xdata(x+self.side)
        self.pt_side.set_ydata(y)
        
        for c in self.patches:
            c.set_xy((x-self.side,y-self.side))
        
        self.x = x
        self.y = y
        self.update_popup_label()
    
    # ======================================================================= #
    def update_side(self, x, y):
        """
            Update circle radius based on DraggablePoint
        """

        self.side = abs(self.x-x)
        dx = self.side*2
        
        for c in self.patches:
            c.set_xy((self.x-self.side,self.y-self.side))
            c.set_height(dx)
            c.set_width(dx)
            
        self.update_popup_label()
        
class Rectangle(Target):
    """
        Drawing Rectangle target shapes on lots of figures
        
        Data fields:
            pt_tr, pt_tl, pt_br, pt_bl: DraggablePoints
            patches:    mpl.patches for patches
            x,y:    center coordinates
            dx,dy:  side length
    """

    # ======================================================================= #
    def __init__(self, popup_target, color, label, x, y, side):
        
        super().__init__(popup_target, color, label)
        
        # save circle position
        self.x = x
        self.y = y
        self.dx = side
        self.dy = side
        
        # place circle at the center of the window
        self.patches = []
        
        # corner points (tr = top right)
        self.pt_tl = DraggablePoint(self,self.update_tl,
                            setx=True,sety=True,color=self.color, marker='s')
        
        self.pt_tr = DraggablePoint(self,self.update_tr,
                            setx=True,sety=True,color=self.color, marker='s')

        self.pt_br = DraggablePoint(self,self.update_br,
                            setx=True,sety=True,color=self.color, marker='s')

        self.pt_bl = DraggablePoint(self,self.update_bl,
                            setx=True,sety=True,color=self.color, marker='s')
        
        self.points = [self.pt_tl, self.pt_tr, self.pt_br, self.pt_bl]
        self.update_popup_label()
        
    # ======================================================================= #
    def draw(self, ax):
        """Add the target to the current axes"""
        
        super().draw(ax)
        
        # draw things
        self.patches.append(patches.Rectangle((self.x-self.dx,self.y-self.dy),
                                            width=self.dx*2,
                                            height=self.dy*2,
                                            fill=False, 
                                            facecolor='none',
                                            lw=1,
                                            ls='-',
                                            edgecolor=self.color))
        ax.add_patch(self.patches[-1])
        self.pt_tr.add_ax(ax, self.x+self.dx, self.y+self.dy)
        self.pt_tl.add_ax(ax, self.x-self.dx, self.y+self.dy)
        self.pt_br.add_ax(ax, self.x+self.dx, self.y-self.dy)
        self.pt_bl.add_ax(ax, self.x-self.dx, self.y-self.dy)
        
    # ======================================================================= #
    def update_popup_label(self):
        """Update popup window label with info on target"""
        
        self.label.config(text='x0 = %.3f\ny0 = %.3f\ndx = %.3f\ndy = %.3f' % \
                (self.x, self.y, self.dx*2, self.dy*2))
        
    # ======================================================================= #
    def update_tr(self, x, y):
        """
            Update top right position based on DraggablePoint
        """
        self.pt_tl.set_ydata(y)
        self.pt_br.set_xdata(x)
        
        ddx = x - int(self.pt_tl.get_xdata())
        ddy = y - int(self.pt_br.get_ydata())
        
        dx = round(ddx/2)
        dy = round(ddy/2)
        
        for c in self.patches:
            c.set_xy((x-ddx,y-ddy))
            c.set_width(ddx)
            c.set_height(ddy)
        
        self.x = x-dx
        self.y = y-dy
        
        self.dx = abs(dx)
        self.dy = abs(dy)
        
        self.update_popup_label()
    
    # ======================================================================= #
    def update_tl(self, x, y):
        """
            Update top left position based on DraggablePoint
        """
        self.pt_tr.set_ydata(y)
        self.pt_bl.set_xdata(x)
        
        ddx = int(self.pt_tr.get_xdata()) - x
        ddy = y- int(self.pt_bl.get_ydata())
        
        dx = round(ddx/2)
        dy = round(ddy/2)
        
        for c in self.patches:
            c.set_xy((x,y-ddy))
            c.set_width(ddx)
            c.set_height(ddy)
        
        self.x = x+dx
        self.y = y-dy
        
        self.dx = abs(dx)
        self.dy = abs(dy)
        
        self.update_popup_label()
        
    # ======================================================================= #
    def update_br(self, x, y):
        """
            Update bottom right position based on DraggablePoint
        """
        self.pt_bl.set_ydata(y)
        self.pt_tr.set_xdata(x)
        
        ddx = x - int(self.pt_bl.get_xdata())
        ddy = int(self.pt_tr.get_ydata()) - y
        
        dx = round(ddx/2)
        dy = round(ddy/2)
        
        for c in self.patches:
            c.set_xy((x-ddx,y))
            c.set_width(ddx)
            c.set_height(ddy)
        
        self.x = x-dx
        self.y = y+dy
        
        self.dx = abs(dx)
        self.dy = abs(dy)
        
        self.update_popup_label()
        
    # ======================================================================= #
    def update_bl(self, x, y):
        """
            Update bottom left position based on DraggablePoint
        """
        self.pt_br.set_ydata(y)
        self.pt_tl.set_xdata(x)
        
        ddx = int((self.pt_br.get_xdata() - x))
        ddy = int((self.pt_tl.get_ydata() - y))
        
        dx = round(ddx/2)
        dy = round(ddy/2)
        
        for c in self.patches:
            c.set_xy((x,y))
            c.set_width(ddx)
            c.set_height(ddy)
        
        self.x = x+dx
        self.y = y+dy
        
        self.dx = abs(dx)
        self.dy = abs(dy)
        
        self.update_popup_label()
                
class Ellipse(Target):
    """
        Drawing ellipse target shapes on lots of figures
        
        Data fields:
            pt_center, pt_radius1, pt_radius2: DraggablePoints
            patches: mpl.patches for patches
            x,y: center coordinates
            r1, r2: radius
    """
    
    # minimum radius able to set
    rmin = 10

    # ======================================================================= #
    def __init__(self, popup_target, color, label, x, y, r1, r2):
        
        super().__init__(popup_target, color, label)
        
        # save circle position
        self.x = x
        self.y = y
        self.r1 = r1
        self.r2 = r2
        self.angle = 0
        
        # place circle at the center of the window
        self.patches = []
        
        # center
        self.pt_center = DraggablePoint(self,self.update_center,
                            setx=True,sety=True,color=self.color, marker='x')
        
        # radius
        self.pt_radius1 = DraggablePoint(self,self.update_radius1,
                            setx=True,sety=True,color=self.color, marker='o')
        
        self.pt_radius2 = DraggablePoint(self,self.update_radius2,
                            setx=True,sety=True,color=self.color, marker='o')
        
        self.points = [self.pt_center, self.pt_radius1, self.pt_radius2]
        self.update_popup_label()
        
    # ======================================================================= #
    def draw(self, ax):
        """Add the target to the current axes"""
        
        super().draw(ax)
        
        # draw things
        self.patches.append(patches.Ellipse((self.x,self.y),
                                     width=self.r1*2, 
                                     height=self.r2*2,
                                     angle=self.angle*180/np.pi,
                                     fill=False, 
                                     facecolor='none',
                                     lw=1,
                                     ls='-',
                                     edgecolor=self.color))
        ax.add_patch(self.patches[-1])
        self.pt_center.add_ax(ax, self.x, self.y)
        self.pt_radius1.add_ax(ax,  self.x+self.r1*np.cos(self.angle), 
                                    self.y+self.r1*np.sin(self.angle))
        self.pt_radius2.add_ax(ax,  self.x+self.r2*np.cos(self.angle+np.pi/2), 
                                    self.y+self.r2*np.sin(self.angle+np.pi/2))

    # ======================================================================= #
    def update_popup_label(self):
        """Update popup window label with info on target"""
        
        self.label.config(text='x0 = %.3f\ny0 = %.3f\nr1 = %.3f\nr2 = %.3f\nangle = %.3f [rad]' % \
                        (self.x, self.y, self.r1, self.r2, self.angle))
        
    # ======================================================================= #
    def update_center(self, x, y):
        """
            Update circle position based on DraggablePoint
        """
        self.pt_radius1.set_xdata(x+self.r1*np.cos(self.angle))
        self.pt_radius1.set_ydata(y+self.r1*np.sin(self.angle))
        self.pt_radius2.set_xdata(x+self.r2*np.cos(self.angle+np.pi/2))
        self.pt_radius2.set_ydata(y+self.r2*np.sin(self.angle+np.pi/2))
        
        for c in self.patches:
            c.set_center((x,y))
        
        self.x = x
        self.y = y
        self.update_popup_label()
    
    # ======================================================================= #
    def update_radius1(self, x, y):
        """
            Update circle radius based on DraggablePoint
        """
        
        # calculate the xy distance
        dx = x-self.x
        dy = y-self.y

        # get the radius 
        self.r1 = np.sqrt(dx**2+dy**2)
        
        # prevent too small of a radius
        if self.r1 < self.rmin:
            self.r1 = self.rmin
            self.pt_radius1.set_xdata(self.x+self.r1*np.cos(self.angle))
            self.pt_radius1.set_ydata(self.y+self.r1*np.sin(self.angle))
        
        # get the angle 
        try:
            a = np.arctan(dy/dx)
        except RuntimeWarning:
            pass
        else:
            self.angle = a
            
        # big angles
        if dx < 0:
            self.angle += np.pi
        
        # set patch
        for c in self.patches:
            c.set_width(self.r1*2)
            c.set_angle(self.angle*180/np.pi)
            
        # set r2
        self.pt_radius2.set_xdata(self.x+self.r2*np.cos(self.angle+np.pi/2))
        self.pt_radius2.set_ydata(self.y+self.r2*np.sin(self.angle+np.pi/2))
            
        self.update_popup_label()
    
    # ======================================================================= #
    def update_radius2(self, x, y):
        """
            Update circle radius based on DraggablePoint
        """

        # calculate the xy distance
        dx = x-self.x
        dy = y-self.y

        # get the radius 
        self.r2 = np.sqrt(dx**2+dy**2)
        
        # prevent too small of a radius
        if self.r2 < self.rmin:
            self.r2 = self.rmin
            self.pt_radius2.set_xdata(self.x+self.r2*np.cos(self.angle+np.pi/2))
            self.pt_radius2.set_ydata(self.y+self.r2*np.sin(self.angle+np.pi/2))
        
        # get the angle 
        try:
            a = np.arctan(dy/dx)+np.pi/2
        except RuntimeWarning:
            pass
        else:
            self.angle = a
        
         # big angles
        if dx > 0:
            self.angle -= np.pi
            
        # set patch
        for c in self.patches:
            c.set_height(self.r2*2)
            c.set_angle(self.angle*180/np.pi)
            
        # set r1
        self.pt_radius1.set_xdata(self.x+self.r1*np.cos(self.angle))
        self.pt_radius1.set_ydata(self.y+self.r1*np.sin(self.angle))            
        self.update_popup_label()
        
class DraggablePoint:

    # http://stackoverflow.com/questions/21654008/matplotlib-drag-overlapping-points-interactively
    # https://stackoverflow.com/questions/28001655/draggable-line-with-draggable-points
    
    lock = None #  only one can be animated at a time
    size=8

    # ======================================================================= #
    def __init__(self, parent, updatefn, setx=True, sety=True, color=None, marker='s'):
        """
            parent: parent object
            points: list of point objects, corresponding to the various axes 
                    the target is drawn in 
            updatefn: funtion which updates the line in the corpatchest way
                updatefn(xdata,ydata)
            x,y: initial point position
            setx,sety: if true, allow setting this parameter
            color: point color
        """
        self.parent = parent
        self.points = []
        self.color = color
        self.marker = marker
            
        self.updatefn = updatefn
        self.setx = setx
        self.sety = sety
        self.press = None
        self.background = None
        
        # trackers for connections
        self.cidpress = []
        self.cidrelease = []
        self.cidmotion = []
        
    # ======================================================================= #
    def add_ax(self, ax, x=None, y=None):
        """Add axis to list of axes"""
        
        self.disconnect()
        
        if x is None:
            x = self.get_xdata()
        if y is None:
            y = self.get_ydata()
            
        self.points.append(ax.plot(x, y, zorder=100, color=self.color, alpha=0.5, 
                        marker=self.marker, markersize=self.size)[0])
        self.points[-1].set_pickradius(self.size)
        
        self.connect()
        
    # ======================================================================= #
    def connect(self):
        """connect to all the events we need"""
        
        # trackers for connections
        self.cidpress = []
        self.cidrelease = []
        self.cidmotion = []
        
        for i,pt in enumerate(self.points):
            self.cidpress.append(pt.figure.canvas.mpl_connect('button_press_event', 
                                partial(self.on_press, id=i)))
                                 
            self.cidrelease.append(pt.figure.canvas.mpl_connect('button_release_event', 
                                self.on_release))
                                
            self.cidmotion.append(pt.figure.canvas.mpl_connect('motion_notify_event', 
                                partial(self.on_motion, id=i)))
    
    # ======================================================================= #
    def on_press(self, event, id):
        
        if event.inaxes != self.points[id].axes: return
        if DraggablePoint.lock is not None: return
        contains, attrd = self.points[id].contains(event)
        if not contains: return
        DraggablePoint.lock = self
        
    # ======================================================================= #
    def on_motion(self, event, id):

        if DraggablePoint.lock is not self: return
        if event.inaxes != self.points[id].axes: return
        
        # get data
        x = event.xdata
        y = event.ydata
        
        # move the point
        if self.setx:   self.set_xdata(x)
        if self.sety:   self.set_ydata(y)

        # update the line
        self.updatefn(x,y)        

    # ======================================================================= #
    def on_release(self, event):
        'on release we reset the press data'
        if DraggablePoint.lock is not self: return
        DraggablePoint.lock = None
        
    # ======================================================================= #
    def disconnect(self):
        'disconnect all the stored connection ids'
        
        for i,pt in enumerate(self.points):
            pt.figure.canvas.mpl_disconnect(self.cidpress[i])
            pt.figure.canvas.mpl_disconnect(self.cidrelease[i])
            pt.figure.canvas.mpl_disconnect(self.cidmotion[i])
                
    # ======================================================================= #
    def get_xdata(self):
        """Get x coordinate"""
        return self.points[0].get_xdata()
            
    # ======================================================================= #
    def get_ydata(self):
        """Get y coordinate"""
        return self.points[0].get_ydata()
            
    # ======================================================================= #
    def remove(self, ax):
        """
            Remove drawn points from the axis.
        """
    
        self.disconnect()
        del_list = []
        
        # remove from draggable points
        for i,line in enumerate(ax.lines):
            for pt in self.points:
                if line is pt:
                    del_list.append(i)
                    self.points.remove(line)
                    del line
        
        # remove from mpl
        for d in del_list:
            ax.lines[d].remove()
        
        self.connect()
            
    # ======================================================================= #
    def set_xdata(self, x):
        """Set x coordinate"""
        for pt in self.points:
            pt.set_xdata(x)    
            
    # ======================================================================= #
    def set_ydata(self, y):
        """Set y coordinate"""
        for pt in self.points:
            pt.set_ydata(y)



# Draw main bccd gui
# Derek Fujimoto
# Sep 2020

from tkinter import *
from tkinter import ttk, filedialog

# set MPL backend
import matplotlib as mpl
mpl.use('TkAgg')

import sys, os, datetime
import matplotlib.pyplot as plt
import numpy as np
import weakref as wref
import subprocess
import yaml

from bccd import __version__, icon_path
from bccd.backend.PltTracker import PltTracker
from bccd.gui.fits_tab import fits_tab
from bccd.gui.popup_target import popup_target
import bccd.backend.colors as colors

# interactive plotting
plt.ion()

__doc__ = """

"""

# =========================================================================== #
class bccd(object):
    """
        Build the mainframe and set up the runloop for the tkinter GUI. 
        
        Data Fields:
            
            mainframe: frame for root
            notebook: notebook for adding files
            tabs: list of fits_tabs objects which have been fetched fits_tabs
            targets: list of popup_target objects
    """
    
    # image fetch locations
    data_remote = ['bnmr@bnmrexp.triumf.ca:/home/bnmr/CCD-Images', 
                   'bnqr@bnqrexp.triumf.ca:/home/bnqr/CCD-Images']
    
    # where to store data on local machine
    data_local = os.path.join(os.environ['HOME'], '.bccd')
    
    # last accessed directory when fetching files
    cwd = os.path.join(os.environ['HOME'], '.bccd')
    
    # rescale pixels flag
    rescale_pixels = True
    
    # ======================================================================= #
    def __init__(self):
        """"""
        
        # plot tracker
        self.plt = PltTracker()
        
        # root 
        root = Tk()
        self.root = root
        root.title("βccd: β-NMR and β-NQR Beamspot Viewer "+\
                   "(version %s)" % __version__)
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)
        
        # hotkeys
        root.bind('<Return>',self.key_return)             
        root.bind('<KP_Enter>',self.key_return)
        root.bind('<Control-Key-Return>',self.key_ctrl_return)      
        root.bind('<Control-Key-KP_Enter>',self.key_ctrl_return)
        root.bind('<Control-Key-1>',lambda x: self.key_ctrl_n(0))
        root.bind('<Control-Key-2>',lambda x: self.key_ctrl_n(1))
        root.bind('<Control-Key-3>',lambda x: self.key_ctrl_n(2))
        root.bind('<Control-Key-4>',lambda x: self.key_ctrl_n(3))
        root.bind('<Control-Key-5>',lambda x: self.key_ctrl_n(4))
        root.bind('<Control-Key-6>',lambda x: self.key_ctrl_n(5))
        root.bind('<Control-Key-7>',lambda x: self.key_ctrl_n(6))
        root.bind('<Control-Key-8>',lambda x: self.key_ctrl_n(7))
        root.bind('<Control-Key-9>',lambda x: self.key_ctrl_n(8))
        root.bind('<Control-Key-0>',lambda x: self.key_ctrl_n(9))
        root.bind('<Control-w>',self.key_ctrl_w)
        root.bind('<Control-o>',self.key_ctrl_o)
        root.bind('<Control-l>',self.key_ctrl_l)
        root.bind('<Control-t>',self.key_ctrl_t)
        
        # styling
        root.option_add('*tearOff', FALSE)
        root.option_add("*Font", colors.font)
        root.option_add("*Background",          colors.background)
        root.option_add("*DisabledBackground",  colors.background)
        root.option_add("*ReadonlyBackground",  colors.readonly)
        root.option_add("*Borderwidth", 2)
        
        # don't change all foregrounds or you will break the filedialog windows
        root.option_add("*Menu*Foreground",     colors.foreground)   
        root.option_add("*Spinbox*Foreground",  colors.foreground)
        root.option_add("*Listbox*Foreground",  colors.foreground)
        root.option_add("*Text*Foreground",     colors.foreground)
        
        root.option_add("*Scrollbar.Background", colors.foreground)
        
        ttk_style = ttk.Style()
        ttk_style.configure('.', font=colors.font, 
                                   background=colors.background, 
                                   foreground=colors.foreground, 
                                   arrowcolor=colors.foreground, 
                                   borderwidth=2)
                                   
        ttk_style.map('.', background=[('disabled', colors.background)], 
                           fieldbackground=[('selected', colors.selected)])
                                         
        ttk_style.configure('TNotebook.Tab', padding=[5, 2])
        ttk_style.configure("TNotebook.Tab", background=colors.background)
        ttk_style.map("TNotebook.Tab", background=[("selected", colors.tab)])
        
        ttk_style.configure("TEntry", foreground=colors.foreground, 
                                     fieldbackground=colors.fieldbackground)
        
        ttk_style.map("TEntry", foreground    =[('active',  colors.foreground), 
                                                ('disabled', colors.disabled)], 
                               fieldbackground=[('active',  colors.fieldbackground), 
                                                ('disabled', colors.disabled), 
                                                ('readonly', colors.readonly)])
                                                                         
        ttk_style.map("TCheckbutton", foreground=[('selected', colors.selected), 
                                                  ('disabled', colors.disabled)], 
                                    indicatorcolor=[('selected', 'green3')])
        ttk_style.map('TCombobox', fieldbackground=[('readonly', colors.background)])
        
        ttk_style.configure('TSpinbox', borderwidth=0, background=colors.background)
        ttk_style.map('TSpinbox', borderwidth=[('selected', 1)])
        
        # icon
        self.set_icon(root)
        
        # event bindings
        root.protocol("WM_DELETE_WINDOW", self.on_closing)
        
        # drawing styles
        self.style = {'linestyle':'None', 
                      'linewidth':mpl.rcParams['lines.linewidth'], 
                      'marker':'.', 
                      'markersize':mpl.rcParams['lines.markersize'], 
                      'capsize':0., 
                      'elinewidth':mpl.rcParams['lines.linewidth'], 
                      'alpha':1., 
                      'fillstyle':'full'}
        
        # main frame
        self.mainframe = ttk.Frame(root, pad=5)
        self.mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
        self.mainframe.columnconfigure(0, weight=1)
        self.mainframe.rowconfigure(0, weight=1)
        
        # Menu bar options ----------------------------------------------------
        root.option_add('*tearOff', FALSE)
        menubar = Menu(root)
        root['menu'] = menubar
        
        # File
        menu_file = Menu(menubar)
        menu_file.add_command(label='Load From Yaml', command=self.load)
        menu_file.add_command(label='Close All Figures', command=self.close_all)
        menu_file.add_command(label='Exit', command=sys.exit)
        menubar.add_cascade(menu=menu_file, label='File')
        
        # Top Notebook --------------------------------------------------------
        noteframe = ttk.Frame(self.mainframe, relief='sunken', pad=5)
        self.notebook = ttk.Notebook(noteframe)
        
        # Buttons -------------------------------------------------------------
        button_add_file = ttk.Button(self.mainframe, text='Add Image', 
                                     command=self.add_file, pad=5)
        button_addlast_file = ttk.Button(self.mainframe, text='Add Last', 
                                     command=self.addlast_file, pad=5)
        button_target = ttk.Button(self.mainframe, text='New Target', 
                                     command=self.addtarget, pad=5)
        
        # gridding
        self.notebook.grid(column=0, row=0, sticky=(N, E, W, S))
        button_target.grid(column=0, row=1, sticky=(E, S))
        button_add_file.grid(column=1, row=1, sticky=(E, S))
        button_addlast_file.grid(column=2, row=1, sticky=(E, S))
        noteframe.grid(column=0, row=0, sticky=(N, E, W, S), columnspan=4, 
                       pady=(0,10))
        noteframe.columnconfigure(0, weight=1)
        noteframe.rowconfigure(0, weight=1)

        # make data directory
        os.makedirs(self.data_local, exist_ok=True)
        
        # intialize tabs list
        self.tabs = []
        
        # intialize targets list
        self.targets = []

        # runloop
        self.root.mainloop()
    
    # ======================================================================= #
    def _add_tab(self, filename):
        
        if len(self.tabs) > 0:
            new_key = max([t.id for t in self.tabs])+1
        else:
            new_key = 0
        
        tab_frame = ttk.Frame(self.notebook, pad=5)
        self.notebook.add(tab_frame, text='Img %d' % (new_key+1))
        
        self.tabs.append(fits_tab(wref.proxy(self), tab_frame, filename, new_key))
        self.notebook.select(len(self.tabs)-1)
        
        # set alpha to prior image value
        try:
            tab = self.tabs[-2]
            alpha = tab.input_objs['alpha'][0].get()
        except KeyError:
            alpha = tab.old_alpha
        except IndexError:
            pass
        else:
            self.tabs[-1].input_objs['alpha'][0].set(alpha)
        
    # ======================================================================= #
    def add_file(self):
        """
            Add tab based on new file
        """
        
        # get data
        self.get_data()
        
        # get filename from browser
        imgfiles = filedialog.askopenfilenames(initialdir=self.cwd,
                                            title='Select CCD Image',
                                            filetypes=(('fits','*.fits'),('All','*')))
        
        if not imgfiles:
            return
        
        # add tab
        for imgfile in imgfiles:                                    
            self.cwd = os.path.split(os.path.abspath(imgfile))[0]            
            self._add_tab(imgfile)
        
    # ======================================================================= #
    def addlast_file(self):
        """
            Add tab based on last modified file
            
        """
        
        # get data
        self.get_data()
        
        # get filenames of all files in directory system and their modification 
        # times
        latest_file = ''
        latest_time = 0
        
        for dirpath, dirnames, filenames in os.walk(self.data_local):
            for f in filenames:
                fname = os.path.join(dirpath,f)
                mod_time = os.stat(fname).st_mtime
                
                if mod_time > latest_time:
                    latest_time = mod_time
                    latest_file = fname                
        
        self.cwd = os.path.split(os.path.abspath(latest_file))[0]
        
        # add tab
        self._add_tab(latest_file)
    
    # ======================================================================= #
    def addtarget(self):
        """
            Add new target to drawn windows
        """
        
        # get number of possible colors
        ncolor = len(list(popup_target.colors.keys()))
        
        # get next target color
        if self.targets:
            
            # get all colors 
            clrs = [t.color for t in self.targets]
            
            # get color indices
            clrs_i = [int(c[1]) for c in clrs]
            
            # find missing in sequence
            missing = [i for i in range(ncolor) if i not in clrs_i]
            
            color = 'C%d' % min(missing)
        else:
            color = 'C0'
            
        self.targets.append(popup_target(self, color))
        
    # ======================================================================= #
    def close_all(self):
        """Close all open figures"""
        plt.close('all')
        self.plt.plots = []
        self.plt.active = 0

    # ====================================================================== #
    def get_data(self):
        """Fetch the images from a remote location"""
        
        for loc in self.data_remote:
            
            # make destination location 
            dest = os.path.join(self.data_local, loc.split(':')[0])
            os.makedirs(dest, exist_ok=True)
            
            # rsync
            print("Fetching data from %s..." % loc)
            subprocess.call(['rsync', '-az', '--min-size=1', os.path.join(loc,'*'), dest])
            
    # ====================================================================== #
    def key_ctrl_l(self,*args):
        """
            Bound to <Control-Key-l>. 
            Open latest file
        """
        self.addlast_file()
    
    # ====================================================================== #
    def key_ctrl_n(self,n,*args):
        """
            Bound to <Control-Key-#>
            Switch to tab n
        """
        try:
            self.notebook.select(n)
        except Exception:
            pass
        
    # ====================================================================== #
    def key_ctrl_o(self,*args):
        """
            Bound to <Control-Key-o>
            Open file browser
        """
        self.add_file()
    
    # ====================================================================== #
    def key_ctrl_t(self,*args):
        """
            Bound to <Control-Key-t>
            New target
        """
        self.addtarget()
        
    # ====================================================================== #
    def key_ctrl_return(self,*args):
        """
            Bound to <Control-Key-Return> and <Control-Key-KP_Enter>
            Add last image
        """
        try:
            idx = self.notebook.index('current')
        except:
            return
        tab = self.tabs[idx]
        tab.draw_new()     
        
    # ====================================================================== #
    def key_ctrl_w(self,*args):
        """
            Bound to <Control-Key-w>. 
            Closes the open tab
        """
        try:
            idx = self.notebook.index('current')
        except:
            return
        self.tabs[idx].close()
        
    # ====================================================================== #
    def key_return(self,*args):
        """
            Bound to <Return> and <KP_Enter>
            Draws in open window
        """
        try:
            idx = self.notebook.index('current')
        except:
            return
        tab = self.tabs[idx]
        tab.draw()
    
    # ====================================================================== #
    def on_closing(self):
        """Excecute this when window is closed: destroy and close all plots."""
        # ~ self.logger.info('Closing all windows.')
        plt.close('all')
        self.root.destroy()
    
    # ====================================================================== #
    def load(self):
        """
            Load images from a bccd yaml file
        """
        
        # get the filename
        filename = filedialog.askopenfilename(initialdir=os.environ['PWD'],
                                    title='Select Save File',
                                    filetypes=(('yaml','*.yaml'),('All','*')))
        with open(filename,'r') as fid:
            data = yaml.safe_load(fid)
            
        # add images as tabs
        for i,val in enumerate(data):
            
            # add tab
            self._add_tab(os.path.join(self.cwd,val['id']))
            tab = self.tabs[-1]
            
            # set tab style
            tab.style.set(val['style'])
            tab.combo_style.event_generate("<<ComboboxSelected>>")
            
            # set tab black
            tab.entry_black.delete(0,END)
            tab.entry_black.insert(0,str(val['black']))
            
            # set tab color
            if val['cmap'][-1] == 'r':    
                color = val['cmap'][:-2]
                tab.input_objs['imap'][0].set(True)
            else:
                color = color = val['cmap']
                tab.input_objs['imap'][0].set(fits_tab.colours[color])
            
            tab.input_objs['cmap'][0].set(color)
            
            # set tab alpha
            tab.input_objs['alpha'][0].set(int(val['alpha']*100))
            
            # set contours
            if 'nlevels' in val.keys():
                tab.input_objs['nlevels'][0].set(val['nlevels'])
                
            # set sigma
            if 'sigma' in val.keys():
                tab.input_objs['sigma'][0].set(val['sigma'])
            
            # draw the image
            if i == 0:
                tab.draw_new()
            else:
                tab.draw()
        
    # ======================================================================= #
    def set_icon(self,window):
        """Set the icon for new windows"""
        try:
            img = PhotoImage(file=icon_path)
            window.tk.call('wm', 'iconphoto', window._w, img)
        except Exception as err:
            print(err)
        



# Track which plots are active, which to draw in. 
# Derek Fujimoto
# Sept 2020

import matplotlib as mpl
import matplotlib.pyplot as plt
import yaml
import os

# =========================================================================== #
class PltTracker(object):
    """
        active:         dictionary, id number of active plot
        plots:          dictionary, list of plots drawn for type
    """
    
    # ======================================================================= #
    def __init__(self):
        
        # track which plots exist
        self.plots = []
        
        # track the active plot 
        self.active = 0
    
    # ======================================================================= #
    def _close_figure(self,event):
        """Remove figure from list"""
        
        # get number and style
        number = event.canvas.figure.number
        
        # disconnect events
        event.canvas.mpl_disconnect(event.canvas.user_close)
        event.canvas.mpl_disconnect(event.canvas.user_active)
        
        # close the winow
        plt.figure(number).clf()
        plt.close(number)
        
        # remove from list 
        self.plots.remove(number)
        
        # reset active
        try:
            self.active = self.plots[-1]
            plt.figure(self.active)
        except IndexError:
            self.active = 0
                        
    # ======================================================================= #
    def _decorator(self,fn,*args,id=None,unique=True,**kwargs):
        """
            Function wrapper
            
            fn: matplotlib function to operate on 
            args: passed to fn
            kwargs: passed to fn
        """
        
        # switch 
        fig = plt.figure(self.active)
        ax = plt.gca()
            
        # clear old objects
        if unique and id is not None:
            self._remove_drawn_object(ax,id)
        
        # run function 
        output = fn(*args,**kwargs)
        
        # track the drawn object
        if id is not None:
            ax.draw_objs.setdefault(id,[]).append((output,kwargs)) 
        
        return output
    
    # ======================================================================= #
    def _remove_drawn_object(self,ax,draw_id):
        """
            Remove an object labelled by draw_id from the figure, based on draw 
            style.
        """
        
        # check if id is present in drawn data
        if draw_id in ax.draw_objs.keys():

            # get item
            for item in ax.draw_objs[draw_id]:
            
                # strip saveas dict
                item = item[0]
            
                # remove line
                try:
                    item[0].remove()
                except TypeError:
                    try:
                        item.remove()
                    except AttributeError:
                        pass
                else: 
                    # remove errorbars
                    if type(item) is mpl.container.ErrorbarContainer:    
                        for i in item[1]:   i.remove()
                        for i in item[2]:   i.remove()
                        del ax.containers[ax.containers.index(item)]
                        
                    # remove lines
                    elif type(item) is mpl.lines.Line2D:
                        del ax.lines[ax.lines.index(item)]
                        
                    # remove images
                    elif type(item) is mpl.image.AxesImage:
                        del ax.image[ax.image.index(item)]
                
                if type(item) is mpl.contour.QuadContourSet:
                    for coll in item.collections: 
                        if coll in ax.collections:
                            del ax.collections[ax.collections.index(coll)]
                    
            plt.draw()
                
            # clear labels
            del ax.draw_objs[draw_id]
        
   # ======================================================================= #
    def _update_active_id(self,event):
        """
            Update the active figure id based on click event.
        """
        number = event.canvas.figure.number
        self.active = number
    
    # ======================================================================= #
    def annotate(self,id,*args,unique=True,**kwargs):
        return self._decorator(plt.annotate,*args,id=id,unique=unique,**kwargs)
    
    # ======================================================================= #
    def axhline(self,id,*args,unique=True,**kwargs):
        return self._decorator(plt.axhline,*args,id=id,unique=unique,**kwargs)
        
    # ======================================================================= #
    def axvline(self,id,*args,unique=True,**kwargs):
        return self._decorator(plt.axvline,*args,id=id,unique=unique,**kwargs)
        
    # ======================================================================= #
    def clf(self):
        """Clear the figure for a given style"""
        out = self._decorator(plt.clf)
        ax = plt.gca()
        ax.draw_objs = {}
        
    # ======================================================================= # 
    def contour(self, id, X, Y, Z, levels=5, unique=True, info=None, **kwargs):
        
        # get active for this style
        active_style = self.active
        
        # make new figure if needed 
        if active_style == 0:   
            self.figure()
            active_style = self.active
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax,id)
        self._remove_drawn_object(ax,'line')
        
        obj = ax.contour(X, Y, Z, levels, **kwargs)
        
        # check input
        if info is None: info = {}
        
        saveas = {'id'              :id,
                  'levels'          :levels,
                  **info,
                  **kwargs
                  }
        
        # save the drawn object to the file
        ax.draw_objs.setdefault(id,[]).append((obj,saveas))
        
        return obj
        
    # ======================================================================= # 
    def errorbar(self, id, x, y, yerr=None, xerr=None, fmt='', ecolor=None, 
                 elinewidth=None, capsize=None, barsabove=False, lolims=False, 
                 uplims=False, xlolims=False, xuplims=False, errorevery=1, 
                 capthick=None, *, data=None, unique=True, annot_label=None,
                 **kwargs):
        """
            Plot data.
            
            annot_label: list of annotations for each point
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active
        
        # make new figure if needed 
        if active_style == 0:   
            self.figure()
            active_style = self.active
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax,id)
        self._remove_drawn_object(ax,'line')
        
        # set up labelling of line objects for mouseover
        if 'label' in kwargs:
            label = kwargs.pop('label')
        else:
            label = ''
        
        # draw in active style 
        obj = plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=fmt, ecolor=ecolor, 
                     elinewidth=elinewidth, capsize=capsize, 
                     barsabove=barsabove, lolims=lolims, uplims=uplims, 
                     xlolims=xlolims, xuplims=xuplims, errorevery=errorevery, 
                     capthick=capthick, data=data, **kwargs)

        # label the line object
        ax.lines[-1].set_label(label)
        
        # set the annotation
        if annot_label is not None: 
            ax.lines[-1].annot_label = annot_label
            
        saveas = {'id'          :id,
                  'fmt'         :fmt,
                  'color'       :color,
                  'ecolor'      :ecolor,
                  'elinewidth'  :elinewidth, 
                  'capsize'     :capsize, 
                  'barsabove'   :barsabove, 
                  'lolims'      :lolims, 
                  'uplims'      :uplims, 
                  'xlolims'     :xlolims,
                  'xuplims'     :xuplims,
                  'errorevery'  :errorevery, 
                  'capthick'    :capthick,
                  **kwargs,
                  }
            
        # save the drawn object to the file
        ax.draw_objs.setdefault(id,[]).append((obj,saveas))
        
        return obj
    
    # ======================================================================= #
    def figure(self,**kwargs):
        """
            Make new figure.
            
            kwargs: keyword arguments to pass to plt.figure
        """
        
        # make figure
        fig = plt.figure(**kwargs)
        
        # make events and save as canvas attribute
        fig.canvas.user_close = fig.canvas.mpl_connect('close_event', self._close_figure)
        fig.canvas.user_active = fig.canvas.mpl_connect('button_press_event', self._update_active_id)
        
        # set window name 
        fig.canvas.set_window_title('Figure %d' % (fig.number))
        
        # update lists
        self.plots.append(fig.number)
        self.active = fig.number
        
        # track drawn objects
        ax = plt.gca()
        if not hasattr(ax,'draw_objs'):
            ax.draw_objs = {}
        
        # setup saveas
        fig.saveas = fig.savefig
        fig.savefig = self.savefig_new
        
        return fig

    # ======================================================================= #
    def gca(self):
        if not self.plots: self.figure()
        return self._decorator(plt.gca)
    
    # ======================================================================= #
    def gcf(self):
        if not self.plots: self.figure()
        return self._decorator(plt.gcf)
    
    # ======================================================================= #
    def imshow(self, id, X, cmap=None, norm=None, aspect=None, interpolation=None, 
                alpha=None, vmin=None, vmax=None, origin=None, extent=None, 
                filternorm=1, filterrad=4.0, resample=None, url=None, data=None,
                unique=True, info=None, **kwargs):
        """
            unique: force only one of this id in the figure
            info: dict of other info to pass to plttracker, save for writing later
        """
        
        
        # get active for this style
        active_style = self.active
        
        # make new figure if needed 
        if active_style == 0:   
            self.figure()
            active_style = self.active
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax,id)
        
        # set up labelling of line objects for mouseover
        if 'label' in kwargs:
            label = kwargs.pop('label')
        else:
            label = ''
        
        # draw in active style 
        obj = plt.imshow(X=X, cmap=cmap, norm=norm, aspect=aspect, 
                interpolation=interpolation, alpha=alpha, vmin=vmin, vmax=vmax, 
                origin=origin, extent=extent, filternorm=filternorm, 
                filterrad=filterrad, resample=resample, url=url, data=data,
                **kwargs)
            
        # check input
        if info is None: info = {}
            
        # save the drawn object to the file
        saveas = {'id'              :id,
                  'cmap'            :cmap,
                  'norm'            :norm,
                  'aspect'          :aspect,
                  'interpolation'   :interpolation,
                  'alpha'           :alpha,
                  'vmin'            :vmin,
                  'vmax'            :vmax,
                  'origin'          :origin,
                  'extent'          :extent,
                  'filternorm'      :filternorm,
                  'filterrad'       :filterrad,
                  'resample'        :resample,
                  'url'             :url,
                  **info,
                  **kwargs
                  }
        
        ax.draw_objs.setdefault(id,[]).append((obj,saveas))
        
        return obj
        
    # ======================================================================= #
    def legend(self,*args,**kwargs):
        self._decorator(plt.legend,*args,**kwargs)
        
    # ======================================================================= #
    def plot(self,id,*args,scalex=True,scaley=True,data=None,unique=True,**kwargs):
        """
            Plot data.
            
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active
        
        # make new figure if needed 
        if active_style not in self.plots:   
            self.figure()
        
        # draw in active style 
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax,id)
        self._remove_drawn_object(ax,'line')
        
        obj = plt.plot(*args,scalex=scalex, scaley=scaley, data=data,**kwargs)
        ax.draw_objs.setdefault(id,[]).append(obj)
        
        return obj

    # ======================================================================= #
    def savefig_new(self,filename):            
        """Save figure, alongside yaml file with figure details"""
        fig = self.gcf()
        
        # get paramters for drawn objs
        draw_objs = self.gca().draw_objs.values()
        objs = []
        for drawo in draw_objs:
            
            for i,draw in enumerate(drawo):
                
                if '.bccd/' in draw[1]['id']:
                    draw[1]['id'] = draw[1]['id'].split('.bccd/')[1]
                
                objs.append(draw[1])
        
        # write text file with details
        yaml_file = os.path.splitext(filename)[0]+'.yaml'
        with open(yaml_file,'w') as fid:    
            fid.write(yaml.dump(objs))
        
        # write figure
        fig.saveas(filename)
        
    # ======================================================================= #
    def text(self,*args,id=None,unique=True,**kwargs):
        return self._decorator(plt.text,*args,id=id,unique=unique,**kwargs)

    # ======================================================================= #
    def tight_layout(self,*args,**kwargs):
        return self._decorator(plt.tight_layout,*args,**kwargs)

    # ======================================================================= #
    def xlabel(self,*args,**kwargs):
        return self._decorator(plt.xlabel,*args,**kwargs)
    
    # ======================================================================= #
    def xlim(self,*args,**kwargs):
        return self._decorator(plt.xlim,*args,**kwargs)
    
    # ======================================================================= #
    def xticks(self,*args,**kwargs):
        return self._decorator(plt.xticks,*args,**kwargs)
    
    # ======================================================================= #
    def ylabel(self,*args,**kwargs):
        return self._decorator(plt.ylabel,*args,**kwargs)
    
    # ======================================================================= #
    def ylim(self,*args,**kwargs):
        return self._decorator(plt.ylim,*args,**kwargs)
    
    # ======================================================================= #
    def yticks(self,*args,**kwargs):
        return self._decorator(plt.yticks,*args,**kwargs)
    
    # ======================================================================= #
    def zlabel(self,*args,**kwargs):
        return self._decorator(plt.zlabel,*args,**kwargs)
    
    # ======================================================================= #
    def zlim(self,*args,**kwargs):
        return self._decorator(plt.zlim,*args,**kwargs)
    
    # ======================================================================= #
    def zticks(self,*args,**kwargs):
        return self._decorator(plt.zticks,*args,**kwargs)
    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bccd",
    version="2.4.0",
    author="Derek Fujimoto",
    author_email="fujimoto@phas.ubc.ca",
    description="BNMR/BNQR Beamspot Image Viewer and Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dfujim/bccd",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=['numpy>=1.19','matplotlib>=3.2.2','pandas>=1.0.5',
                      'scipy>=1.5.1','scikit-image>=0.17.2','astropy>=3.2.1'],
    package_data={'': ['./images']},
    include_package_data=True,
)



# Fitting functions
# Derek Fujimoto
# Oct 2019

import numpy as np

def gaussian2D(x,y,x0,y0,sigmax,sigmay,amp,theta=0):
    """Gaussian in 2D - from wikipedia"""
    
    ct2 = np.cos(theta)**2
    st2 = np.sin(theta)**2
    s2t = np.sin(2*theta)
    
    sx = sigmax**2
    sy = sigmay**2
    
    a = 0.5*(ct2/sx + st2/sy)
    b = 0.25*s2t*(-1/sx + 1/sy)
    c = 0.5*(st2/sx + ct2/sy)
    
    return amp*np.exp(-(a*np.square(x-x0) + 2*b*(x-x0)*(y-y0) + c*np.square(y-y0)))

